/*
 -- Multicore --
 !! FM synth is being run by interpolators !!
 and sequenced by a thread on core1
 it could also be sequenced by an interrupt
 Two lanes are used for the main frequency and the modulation frequency
 And one more lane is used for the amplitude envelope 

 */

#include "hardware/gpio.h"
#include "hardware/timer.h"
#include <hardware/irq.h>
#include <hardware/adc.h>
#include <hardware/pwm.h>
#include "hardware/interp.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include <pico/divider.h>
#include "hardware/sync.h"
#include "stdlib.h"
#include "math.h"

// ==========================================
// === set up interpolator
// ==========================================

// Interpolator0-lane1 (FM modulation frequency) setup to:
//  add accum1 + base1 amd store in accum0 (result1 to accum1) (add raw -- no shift)
//  right-shift accum1 23 bits and mask to bits 8:1 (zero low bit for short pointer) 
//  add shifted/masked accum0 to base2 and read result2 as table position output to interp1

// Interpolator0-lane0 (AM modulation amplitude) setup to:
//  add accum0 + base0 amd store in accum0 (result0 to accum0) (add raw -- no shift)

// Interpolator1-lane1 (main oscillator) setup to:
//  add accum1 + base1 amd store in accum0 (result1 to accum1) (add raw -- no shift)
//  right-shift accum1 23 bits and mask to bits 8:1 (zero low bit for short pointer)
//  add shifted/masked accum0 to base2 and read result2 as table position output to PWM

void interpolator_for_dds(void){
    // ===
  // --- interp0 - lane 1 - DDS
  interp_config cfg0 = interp_default_config();
  interp_set_config(interp0, 1, &cfg0);
  // base1 + accum1 -> accum1
  //full 32-bit integer add back to accum0
  interp_config_set_add_raw (&cfg0, true) ;
  // (accum0>>23) & 0x1fe
  // shift and mask for index gen assuming
  // word-aligned, short, sine table
  // actual sine-table pointer will be read from 
  // result2 = ((accum1>>23) & 0x1fe) + base2
  interp_config_set_shift (&cfg0, 23) ;
  interp_config_set_mask(&cfg0, 1, 8);
  // update config
  interp_set_config(interp0, 1, &cfg0);

    // ===
  // --- interp0 - lane 0 - envelope decay
  interp_config cfg00 = interp_default_config();
  interp_set_config(interp0, 0, &cfg00);
  // base1 + accum1 -> accum1
  //full 32-bit integer add back to accum0
  interp_config_set_add_raw(&cfg00, true) ;
  //turn off all bits in shifter
  interp_config_set_shift (&cfg00, 31) ;
  interp_config_set_mask(&cfg00, 1, 1);
  // update config
  interp_set_config(interp0, 0, &cfg00);

   // ===
  // --- interp1 - lane 1 - DDS
  interp_config cfg1 = interp_default_config();
  interp_set_config(interp1, 1, &cfg1);
  // base1 + accum1 -> accum1
  //full 32-bit integer add back to accum0
  interp_config_set_add_raw (&cfg1, true) ;
  // (accum0>>23) & 0x1fe
  // shift and mask for index gen assuming
  // word-aligned, short, sine table
  // actual sine-table pointer will be read from 
  // result2 = ((accum1>>23) & 0x1fe) + base2
  interp_config_set_shift (&cfg1, 23) ;
  interp_config_set_mask(&cfg1, 1, 8);
  // update config
  interp_set_config(interp1, 1, &cfg1);
}

// ==========================================
// === set up pwm
// ==========================================
uint pwm_slice_num,  pwm_chan_num ;
void pwm_setup(int pwm_pin){
  gpio_init(pwm_pin) ;
  gpio_set_function(pwm_pin, GPIO_FUNC_PWM) ;
  pwm_slice_num = pwm_gpio_to_slice_num(pwm_pin) ;
  // full speed clock => 256 counts in 2 uSec
  pwm_set_clkdiv(pwm_slice_num, 1.0f);
  pwm_set_clkdiv_mode(pwm_slice_num, PWM_DIV_FREE_RUNNING) ;
  // max count 256
  pwm_set_wrap(pwm_slice_num, 256);
  // initial for testing to half
  pwm_chan_num = pwm_gpio_to_channel(pwm_pin);
  pwm_set_chan_level(pwm_slice_num, pwm_chan_num , 128) ;
  pwm_set_enabled(pwm_slice_num, 1);
}

// 1/Fs in microseconds
volatile int alarm_period = 25 ;
// DDS variables
unsigned int increment0, increment1, fm_depth=20 ; //20;
int dk_rate = -5 ; //-4 ;
short sine_table[256], mod_sine_table[256] ;
float Fs, Fout, Fmod ;
short mod_wave ; 
int amp, input_amp = 500000 ;

// ==========================================
// === set up timer ISR NOT used in this pgm
// ==========================================
// === timer alarm ========================
// !! modifiying alarm zero trashes the cpu 
//        and causes LED  4 long - 4 short
// !! DO NOT USE alarm 0
/*
#define ALARM_NUM 1
#define ALARM_IRQ TIMER_IRQ_1
// ISR interval will be 10 uSec
//
// the actual ISR
static void alarm_irq(void) {
    // mark ISR entry
    //gpio_put(2,1);
    // Clear the alarm irq
    hw_clear_bits(&timer_hw->intr, 1u << ALARM_NUM);
    // arm the next interrupt
    // Write the lower 32 bits of the target time to the alarm to arm it
    timer_hw->alarm[ALARM_NUM] = timer_hw->timerawl + alarm_period ;
    //
    // === do DDS the usual way; ISR time 700 nS 
    //accum += increment ;
    //pwm_set_chan_level(pwm_slice_num, pwm_chan_num , sine_table[accum>>24]) ;

    // === interpolator version ; ISR time 600 nS
    // POP2 contents of pointer in result2 to pwm period 
    // Reading the 'pop' registers also clock the state of the interpolator
    pwm_set_chan_level(pwm_slice_num, pwm_chan_num , *(short*)(interp0->pop[2])) ;

    // mark ISR exit
    //gpio_put(2,0);
}
// set up the timer alarm ISR
static void alarm_in_us(uint32_t delay_us) {
    // Enable the interrupt for our alarm (the timer outputs 4 alarm irqs)
    hw_set_bits(&timer_hw->inte, 1u << ALARM_NUM);
    // Set irq handler for alarm irq
    irq_set_exclusive_handler(ALARM_IRQ, alarm_irq);
    // Enable the alarm irq
    irq_set_enabled(ALARM_IRQ, true);
    // Enable interrupt in block and at processor
    // Alarm is only 32 bits 
    uint64_t target = timer_hw->timerawl + delay_us;
    // Write the lower 32 bits of the target time to the alarm which
    // will arm it
    timer_hw->alarm[ALARM_NUM] = (uint32_t) target;   
}
*/

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// ==================================================
// === user's serial input thread
// ==================================================
// 
static PT_THREAD (protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
    
    // DDS frequenies
    Fout = 400  ;
    Fmod = 550 ;
    // convert alarm period in uSEc to rate
    Fs = 1.0/((float)alarm_period*1e-6) ;
    // increment = Fout/Fs * 2^32
    // the main center freq
    increment1 = Fout * pow(2,32)/ Fs ;

    // the modulation freq
    increment0 = Fmod * pow(2,32)/ Fs ;

    // data structure for interval timer
    PT_INTERVAL_INIT() ;
    //
    
    while(1) {
      
      // == Fout and Fmod are in Hz
      // == fm_depth is an integer: 0 for VERY little modulation, range up to around 20
      sprintf(pt_serial_out_buffer, "\n\rinput main, mod frequency, fm_depth, dk_rate: ");
      serial_write ;
      serial_read ;
      // convert input string to number
      sscanf(pt_serial_in_buffer,"%f %f %d %d", &Fout, &Fmod, &fm_depth, &dk_rate) ;
      // increment = Fout/Fs * 2^32
      increment1 = Fout * pow(2,32 )/ Fs ;
      increment0 = Fmod * pow(2,32 )/ Fs ;
      // set accum0 0 the initial amplitude
      input_amp = 500000 ;
      // make sure dk_rate is negative
      if (dk_rate >= 0) dk_rate = -dk_rate ;

      PT_YIELD_INTERVAL(1000) ;
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    
    // set up LED p25 to blink
    gpio_init(25) ;	
    gpio_set_dir(25, GPIO_OUT) ;
    gpio_put(25, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;
     
    while(1) {
      // yield time 0.1 second
      PT_YIELD_INTERVAL(200000) ;

      // toggle the LED on PICO
      LED_state = LED_state? false : true ;
      gpio_put(25, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === fast_ddds thread -- RUNNING on core 1
// ==================================================
// 
static PT_THREAD (protothread_fast_dds(struct pt *pt))
{
  PT_BEGIN(pt);
    
    // data structure for interval timer
    PT_INTERVAL_INIT() ;

    // init interpolator connections
    interpolator_for_dds();

    // set accum0 0 the initial amplitude
    interp0->accum[0] = input_amp ;
    // decay per sample
    interp0->base[0] = dk_rate ;


    // set base2 to sine_table start addresses
    interp0->base[2] =(int) mod_sine_table ;
    interp1->base[2] =(int) mod_sine_table ;
    //
    // start pwm on pin3
    #define pwm_dds 3
    pwm_setup(pwm_dds);

    // testing
    //gpio_init(2) ;	
    //gpio_set_dir(2, GPIO_OUT) ;
    //gpio_put(2, true);

    while(1) {
      //
     // set rate to 1/alarm_period
     PT_YIELD_INTERVAL(alarm_period) ;
      // start a burst on new data
      if(input_amp > 0) {
          // init the amplitude
          interp0->accum[0] = input_amp ;
          // int the linear dacay rate
          interp0->base[0] = dk_rate ;
          // reset the new data flag
          input_amp = 0 ;
          // phase lock the main frequency
          interp1->accum[1] = 0 ;
      }
      // play the burst as long as the amplitude is positive
      // as it decays linearly
      if ((int)(interp0->accum[0]) > 0) {
        // set dds modulation freq
        interp0->base[1] = increment0 ;
        // set dds main freq (use peek[2] because we need to read another value)
        interp1->base[1] = increment1 + (((int)*(short*)(interp0->peek[2]))<<fm_depth) ;
        // read contunuous FM wave and update interp1 using pop[2]
        mod_wave = *(short*)(interp1->pop[2]) ;
        // read the instataneous amp and update interp0
        amp = (int)(interp0->pop[0]) ;
        // amplitide modulate and shift to the correct range for PWM
        mod_wave = ((long long)mod_wave * (long long)amp) >> 19 ;

        // write final result  to 8-bit PWM    
        //pwm_set_chan_level(pwm_slice_num, pwm_chan_num, amp>>8) ;
        pwm_set_chan_level(pwm_slice_num, pwm_chan_num, mod_wave + 128) ;
      }
      else {
          // turn PWM off
        pwm_set_chan_level(pwm_slice_num, pwm_chan_num, 128) ;
      }
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_fast_dds);
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us

  // dds table
  int i = 0 ;
  while (i<256) {
    sine_table[i] = (short) (127 * sin(2*3.1416*i/256) + 128) ;
    // modulation has no offset
    mod_sine_table[i] = (short) (127 * sin(2*3.1416*i/256)) ;
    i++ ;
  }

  //start timer alarm IRQ
  // In THIS version, core1 takes the place of n ISR
  // toggle a pin for timing
  //gpio_init(2) ;	
  //gpio_set_dir(2, GPIO_OUT) ;
  //alarm_in_us(alarm_period);

  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");
     
  // === add threads ======================
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ===================
  // for core 0
  pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////